#include <v_lib.h>
#include "eval.h"
#include "cspace.h"
#include "messages.h"
#include "istrings.h"

V_LOCAL vp_stack_t	image_cache = TYPED_STACK_EMPTY_INITIALIZER;

V_LOCAL void
rgbFromCie(vec3_t rgb_ret,vec3_t cie)
{
	static matrix3_t cie2rgb_transform={
		{2.739,-1.145,-0.424},
		{-1.11,2.029,0.033},
		{0.138,-0.333,1.105}};
	mat3VecMul(rgb_ret,cie2rgb_transform,cie);
}

V_LOCAL void
cieFromRgb(vec3_t cie_ret,vec3_t rgb)
{
	static matrix3_t rgb2cie_transform={
        {0.477,0.298,0.174},
        {0.261,0.653,0.081},
        {0.019,0.160,0.907}};
	mat3VecMul(cie_ret,rgb2cie_transform,rgb);
}

V_LOCAL void
ctransCieFromRgb(matrix_p cie[],matrix_p rgb[])
{
	u32_t i, j;
	vec3_t rgbv, ciev;
	for(i = 0;i < 3;i++)
	{
		if(cie[i])
			memFree(cie[i]);
		cie[i] = matAlloc(rgb[0]->rows,rgb[0]->cols);
	}
	for(i = 0;i < rgb[0]->rows;i++)
	{
		for(j = 0;j < rgb[0]->cols;j++)
		{
			rgbv[0] = rgb[0]->cell[i][j];
			rgbv[1] = rgb[1]->cell[i][j];
			rgbv[2] = rgb[2]->cell[i][j];
			cieFromRgb(ciev,rgbv);
			cie[0]->cell[i][j] = ciev[0];
			cie[1]->cell[i][j] = ciev[1];
			cie[2]->cell[i][j] = ciev[2];
		}
	}
}

V_LOCAL void
ctransRgbFromCie(matrix_p rgb[],matrix_p cie[])
{
	u32_t i, j;
	vec3_t rgbv, ciev;
	for(i = 0;i < 3;i++)
	{
		if(rgb[i])
			memFree(rgb[i]);
		rgb[i] = matAlloc(cie[0]->rows,cie[0]->cols);
	}
	for(i = 0;i < cie[0]->rows;i++)
	{
		for(j = 0;j < cie[0]->cols;j++)
		{
			ciev[0] = cie[0]->cell[i][j];
			ciev[1] = cie[1]->cell[i][j];
			ciev[2] = cie[2]->cell[i][j];
			rgbFromCie(rgbv,ciev);
			rgb[0]->cell[i][j] = rgbv[0];
			rgb[1]->cell[i][j] = rgbv[1];
			rgb[2]->cell[i][j] = rgbv[2];
		}
	}
}

V_LOCAL void
ctransHsvFromRgb(matrix_p hsv[],matrix_p rgb[])
{
	u32_t i, j;
	vec3_t rgbv, hsvv;
	for(i = 0;i < 3;i++)
	{
		if(hsv[i])
			memFree(hsv[i]);
		hsv[i] = matAlloc(rgb[0]->rows,rgb[0]->cols);
	}
	for(i = 0;i < rgb[0]->rows;i++)
	{
		for(j = 0;j < rgb[0]->cols;j++)
		{
			rgbv[0] = rgb[0]->cell[i][j];
			rgbv[1] = rgb[1]->cell[i][j];
			rgbv[2] = rgb[2]->cell[i][j];
			hsvFromRgb(hsvv,rgbv);
			hsv[0]->cell[i][j] = hsvv[0];
			hsv[1]->cell[i][j] = hsvv[1];
			hsv[2]->cell[i][j] = hsvv[2];
		}
	}
}

V_LOCAL void
ctransRgbFromHsv(matrix_p rgb[],matrix_p hsv[])
{
	u32_t i, j;
	vec3_t rgbv, hsvv;
	for(i = 0;i < 3;i++)
	{
		if(rgb[i])
			memFree(rgb[i]);
		rgb[i] = matAlloc(hsv[0]->rows,hsv[0]->cols);
	}
	for(i = 0;i < hsv[0]->rows;i++)
	{
		for(j = 0;j < hsv[0]->cols;j++)
		{
			hsvv[0] = hsv[0]->cell[i][j];
			hsvv[1] = hsv[1]->cell[i][j];
			hsvv[2] = hsv[2]->cell[i][j];
			rgbFromHsv(rgbv,hsvv);
			rgb[0]->cell[i][j] = rgbv[0];
			rgb[1]->cell[i][j] = rgbv[1];
			rgb[2]->cell[i][j] = rgbv[2];
		}
	}
}

V_LOCAL cspace_desc_t cspaces[] = 
{
	{
		&ipl_key_rgb,
		CSPACE_RGB_OFFSET,
		3,
		{&ipl_key_red,&ipl_key_green,&ipl_key_blue,NULL},
		NULL,
		NULL,
		NULL
	},
	{
		&ipl_key_rgba,
		CSPACE_RGB_OFFSET,
		4,
		{&ipl_key_red,&ipl_key_green,&ipl_key_blue,&ipl_key_alpha},
		NULL,
		NULL,
		NULL
	},
	{
		&ipl_key_cie,
		CSPACE_CIE_OFFSET,
		3,
		{&ipl_key_cieX,&ipl_key_cieY,&ipl_key_cieZ,NULL},
		NULL,
		ctransRgbFromCie,
		ctransCieFromRgb
	},
	{
		&ipl_key_hsv,
		CSPACE_HSV_OFFSET,
		3,
		{&ipl_key_hue,&ipl_key_saturation,&ipl_key_value,NULL},
		NULL,
		ctransRgbFromHsv,
		ctransHsvFromRgb
	},
	{NULL,0,0,{NULL,NULL,NULL,NULL},NULL,NULL,NULL}
};

V_LOCAL cspace_desc_p
cspaceFindDesc(char *key)
{
	cspace_desc_p head = cspaces;
	for(;head->key_ptr;head++)
		if(STR_EQ(*head->key_ptr,key))
			return head;
	return NULL;
}

V_LOCAL vboolean
inputResizePic(image_input_p pic,eval_state_p state)
{
	u32_t i;
	matrix_p m;
	vhandle_t resampler;
	u32_t old_width = pic->spaces[CSPACE_RGB_OFFSET][0]->cols;
	u32_t old_height = pic->spaces[CSPACE_RGB_OFFSET][0]->rows;
	u32_t target_width = old_width;
	u32_t target_height = old_height;
	if(state->custom_width)
		target_width = state->custom_width;
	if(state->custom_height)
		target_height = state->custom_height;
	if(state->input_square)
	{
		u32_t max_size = (target_width > target_height) ? target_width : target_height;
		target_width = target_height = max_size;
	}
	if(state->input_powers2)
	{
		NOT_LESS_POWER_TWO(target_width,target_width,u32_t);
		NOT_LESS_POWER_TWO(target_height,target_height,u32_t);
	}
	if(target_height != old_height || target_width != old_width)
	{
		resampler = matCreateResampleContext(old_height,old_width,
			target_height,target_width);
		if(resampler == LIBV_INVALID_HANDLE)
			return false;
		for(i = 0;i < IPL_MAX_COLOR_CHANNELS;i++)
		{
			if(!pic->spaces[CSPACE_RGB_OFFSET][i])
				continue;
			m = matAlloc(target_height,target_width);
			matResample(resampler,m,pic->spaces[CSPACE_RGB_OFFSET][i]);
			memFree(pic->spaces[CSPACE_RGB_OFFSET][i]);
			pic->spaces[CSPACE_RGB_OFFSET][i] = m;
		}
		matDestroyResampleContext(resampler);
	}
	return true;
}

V_LOCAL image_input_p
inputFindCachePic(eval_state_p state,char *image_name)
{
	CALLER_NAME(inputFindCachePic)
	u32_t i, cur_width, cur_height;
	image_input_p pic, *pics;
	pics = (image_input_p*)vpsGetStorage(&image_cache);
	for(i = 0;i < vpsCount(&image_cache);i++)
	{
		pic = pics[i];
		if(STR_EQ(pic->id,image_name))
		{
			cur_width = pic->spaces[CSPACE_RGB_OFFSET][i]->cols;
			cur_height = pic->spaces[CSPACE_RGB_OFFSET][i]->rows;
			if(state->custom_width && state->custom_width != cur_width)
			{
				WARNING_GENERIC(ipl_size_mismatch);
				continue;
			}
			if(state->custom_height && state->custom_height != cur_height)
			{
				WARNING_GENERIC(ipl_size_mismatch);
				continue;
			}
			if(state->input_square && cur_width != cur_height)
			{
				WARNING_GENERIC(ipl_not_square);
				continue;
			}
			if(state->input_powers2 && 
				(!IS_POWER_TWO(cur_width) || IS_POWER_TWO(cur_height)))
			{
				WARNING_GENERIC(ipl_not_powers2);
				continue;
			}
			return pic;
		}
	}
	pic = (image_input_p)memAllocZ(sizeof(image_input_t));
	V_ASSERT(pic);
	pic->id = strDup(image_name);
	if(!rgbaLoadChannels(pic->id,
		&pic->spaces[CSPACE_RGB_OFFSET][0],
		&pic->spaces[CSPACE_RGB_OFFSET][1],
		&pic->spaces[CSPACE_RGB_OFFSET][2],
		&pic->spaces[CSPACE_RGB_OFFSET][3]) ||
		!inputResizePic(pic,state))
	{
		memFree(pic->id);
		memFree(pic);
		return NULL;
	}
	vpsPush(&image_cache,pic);
	return pic;
}

V_LOCAL vboolean
cspaceFindChannel(char *id,cspace_desc_p *desc_ret,u32_t *ch_offset_ret)
{
	u32_t i;
	cspace_desc_p head = cspaces;
	for(;head->key_ptr;head++)
	{
		for(i = 0;i < IPL_MAX_COLOR_CHANNELS;i++)
		{
			if(head->channel_names[i] && STR_EQ(id,*head->channel_names[i]))
			{
				*desc_ret = head;
				*ch_offset_ret = i;
				return true;
			}
		}
	}
	return false;
}

matrix_p
cspaceGetMatrix(eval_state_p state,char *image_name,char *channel_id)
{
	CALLER_NAME(inputGetMatrix)
	u32_t ch_offset;
	cspace_desc_p cdesc;
	image_input_p pic = inputFindCachePic(state,image_name);
	if(!pic)
		return NULL;
	/* found / loaded; check for non-rgba matix */
	if(!cspaceFindChannel(channel_id,&cdesc,&ch_offset))
	{
		ERROR_GENERIC_2ARG(ipl_unknown_channel,channel_id);
		return NULL;
	}
    if(cdesc->cspace_offset != CSPACE_RGB_OFFSET)
    {
    	if(!pic->spaces[cdesc->cspace_offset][ch_offset])
	    	(*cdesc->native_from_rgba)(pic->spaces[cdesc->cspace_offset],
    			pic->spaces[CSPACE_RGB_OFFSET]);
    }
	return pic->spaces[cdesc->cspace_offset][ch_offset];
}

void
cspaceCollectGarbage(void)
{
	u32_t i, j;
	image_input_p ptr;
	while(vpsPop(&image_cache,(void_p*)&ptr))
	{
		for(i = 0;i < IPL_MAX_COLOR_SPACES;i++)
			for(j = 0;j < IPL_MAX_COLOR_CHANNELS;j++)
				if(ptr->spaces[i][j])
					memFree(ptr->spaces[i][j]);
		memFree(ptr->id);
		memFree(ptr);
	}
	vpsClear(&image_cache);
}

/*
 * evalOutput()
 
	output "out.tga" RGBA <red> <green> <blue> <alpha>

 * Assumes the keyword already read in.
 */
vboolean
cspaceEvalOutput(vhandle_t script)
{
	CALLER_NAME(evalOutput)
	u32_t i;
	arg_p next_channel;
	char file_name[_MAX_PATH];
	cspace_desc_p cdesc;
	matrix_p *cptr, native[IPL_MAX_COLOR_CHANNELS];
	matrix_p rgba[4] = {NULL,NULL,NULL,NULL};
	if(!lexCopyNextToken(script,file_name,_MAX_PATH))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	if(!lexNext(script))
	{
		SYN_UNEXPECTED_EOF(script);
		return false;
	}
	cdesc = cspaceFindDesc(lexLast(script));
	if(!cdesc)
	{
		PARSE_ERROR(script,ipl_wrong_colorspace);
		return false;
	}
	next_channel = evalNextArgument(script,ARG_MATRIX);
	if(!next_channel)
		return false;
	native[0] = next_channel->data.matrix;
	for(i = 1;i < cdesc->components;i++)
	{
		next_channel = evalNextArgument(script,ARG_MATRIX);
		if(!next_channel)
			return false;
		native[i] = next_channel->data.matrix;
		if(!MAT_SAME_SIZE(native[0],native[i]))
		{
			PARSE_ERROR(script,ipl_size_mismatch);
			return false;
		}
	}
	for(;i < IPL_MAX_COLOR_CHANNELS;i++)
		native[i] = NULL;
	if(cdesc->cspace_offset != CSPACE_RGB_OFFSET)
	{
		(*cdesc->rgba_from_native)(rgba,native);
		cptr = rgba;
	}
	else
		cptr = native;
	rgbaSaveChannels(file_name,cptr[0],cptr[1],cptr[2],cptr[3]);
	for(i = 0;i < 4;i++)
	{
		if(rgba[i])
			memFree(rgba[i]);
	}
	return true;
}
